require 'libglade2'

class StorageEntity
	ICON = nil

	attr :row
	attr :data

	def initialize(data, parent = nil)
		storage_initialize(data, parent)
	end

	def storage_initialize(data, parent = nil, sibling = nil)
		@parent = parent
		@data = data
		@row = @data.insert_before(parent, sibling)
		@row[0] = self
	end

	def cleanup
		nil
	end

	def drop
		cleanup
		@data.remove(@row)
	end

	def give(entity)
		false
	end

	def text
		"--NOT IMPLEMENTED--"
	end

	def text_fg
		nil
	end
	
	def act(panel, builder)
		nil
	end

	def opt_panel
		nil
	end

	def prep_opt_panel(panel, builder)
		nil
	end

	def icon
		self.class::ICON
	end

	def new_child
		@data.append(@row)
	end
end

class DriveEntity < StorageEntity
	ICON = Gdk::Pixbuf.new("disk.png")
	ICON_LVM = Gdk::Pixbuf.new("disk_in_lvm.png")
	ICON_MOUNTED = Gdk::Pixbuf.new("disk_mounted.png")
	FREE_ICON = Gdk::Pixbuf.new("disk_new.png")

	attr :name
	attr :size
	attr_accessor :used

	def initialize(data, name, size)
		super(data)
		@name = name
		@size = size
		@next_part = 0

		FreeSpace.new(self)
	end

	def return_space(amount, iter)
		until iter.nil? or iter[0].is_a? FreeSpace
			iter = nil unless iter.next!
		end

		if iter.nil?
			FreeSpace.new(self, @size)
		else
			iter[0].amount += @size
		end
	end

	def icon
		return self.class::ICON if used.nil?
		return self.class::ICON_LVM if used.is_a? VolGroupEntity
		return self.class::ICON_MOUNTED
	end

	def free_icon
		FREE_ICON
	end

	def drop
		super
		self.used = nil
	end

	def mkpart(size, free_row, ignore)
		@next_part += 1
		PartitionEntity.new(self, @next_part, size, free_row)
	end

	def used=(other)
		transition_used if @used.nil? and not other.nil?
		transition_unused if other.nil? and not @used.nil?
		@used.release(self) if @used
		@used = other
	end

	def transition_used
		arr = []

		return unless @row.has_child?
		iter = @data.get_iter(@row.path)
		iter = iter.first_child

		test = true
		while test
			arr << iter[0]
			test = iter.next!
		end

		arr.each{ |x| x.drop }

		if @row.has_child?
			@data.get_iter(@row.path).first_child[0].drop
		end
	end

	def transition_unused
			FreeSpace.new(self, @size)
	end

	def text
		"#{@name} (#{@size}GiB)"
	end

	def opt_panel
		if used
			"part_delete_recall"
		else
			nil
		end
	end

	def prep_opt_panel(panel, builder)
		return nil unless used
		builder.get_object("part_make_name").show
		r_button = builder.get_object("part_recall_button")
		d_button = builder.get_object("part_delete_button")
		label = builder.get_object("part_currently_used_label")

		d_button.hide
		r_button.show
		label.text = "Currently used by #{used.name}"
	end
end

class FreeSpace < StorageEntity
	attr_accessor :amount
	attr :icon

	def initialize(drive, amount = nil)
		amount = drive.size unless amount
		storage_initialize(drive.data, drive.row)

		@drive = drive
		@amount = amount
		@icon = drive.free_icon
	end

	def amount=(amount)
		@amount = amount
		self.drop if @amount == 0
	end

	def text
		"#@amount GiB Free"
	end

	def opt_panel
		"part_make_frame"
	end

	def prep_opt_panel(panel, builder)
		builder.get_object("part_make_name").hide
		do_prep_opt_panel(panel, builder)
	end

	def do_prep_opt_panel(panel, builder)
		bounds = builder.get_object("part_make_size_spinner").adjustment
		label = builder.get_object("part_currently_used_label")
		button = builder.get_object("part_recall_button")
		d_button = builder.get_object("part_delete_button")

		button.show
		d_button.show
		label.show

		bounds.lower = 0
		bounds.upper = @amount
		bounds.value = @amount
	end

	def act(panel, builder)
		size = builder.get_object("part_make_size_spinner").adjustment.value.to_i
		name = builder.get_object("part_name").text
		@drive.mkpart(size, @row, name)
		@amount -= size

		self.drop if @amount == 0
	end

	def text_fg
		"gray"
	end
end

class VGFreeSpace < FreeSpace
	def prep_opt_panel(panel, builder)
		builder.get_object("part_make_name").show
		do_prep_opt_panel(panel, builder)
	end
end

class PartitionEntity < DriveEntity
	ICON = Gdk::Pixbuf.new("disk.png")
	ICON_LVM = Gdk::Pixbuf.new("disk_in_lvm.png")
	ICON_MOUNTED = Gdk::Pixbuf.new("disk_mounted.png")

	attr :number

	def initialize(parent, number, size, sibling = nil)
		storage_initialize(parent.data, parent.row, sibling)
		@number = number
		@size = size
		@name = "#{parent.name}#{number}"
	end

	def transition_used
		nil
	end

	def transition_unused
		nil
	end

	def cleanup
		@parent[0].return_space(@size, @data.get_iter(@row.path))
	end

	def opt_panel
		"part_delete_recall"
	end

	def prep_opt_panel(panel, builder)
		builder.get_object("part_make_name").show
		label = builder.get_object("part_currently_used_label")
		button = builder.get_object("part_recall_button")

		if used
			button.show
			label.show
			label.text = "Currently used by #{used.name}"
		else
			label.hide
			button.hide
		end
	end
end

class VGPartitionEntity < PartitionEntity
	ICON = Gdk::Pixbuf.new("lvm-vg.png")
	ICON_MOUNTED = Gdk::Pixbuf.new("lvm-vg_mounted.png")
end

class MountpointEntity < StorageEntity
	ICON = Gdk::Pixbuf.new("folder.png")
	ICON_MOUNTED = Gdk::Pixbuf.new("folder_mounted.png")

	attr_accessor :fstype
	attr :mounts
	attr :name

	def initialize(data, name, parent = nil)
		storage_initialize(data, parent)
		@name = name
	end

	def icon
		return ICON unless @mounts
		return ICON_MOUNTED
	end

	def opt_panel
		return "mountpoint_info" if @mounts
		return nil if @name == "/"
		return "go_to_mountpoint_box"
	end

	def prep_opt_panel(panel, builder)
		return unless @mounts

		builder.get_object("mountpoint_label").text = @mounts.name
		builder.get_object("fstype").active = 0
	end

	def give(drive)
		return false unless drive.is_a? DriveEntity
		return false if drive.used
		return false if @mounts

		drive.used = self
		@mounts = drive
	end

	def release(drive)
		return unless @mounts == drive
		@mounts = nil
	end

	def text
		@name
	end
end

class VolGroupEntity < StorageEntity
	attr :name
	ICON = Gdk::Pixbuf.new("lvm.png")
	FREE_ICON = Gdk::Pixbuf.new("lvm-vg_new.png")
	PART_ICON = Gdk::Pixbuf.new("lvm-vg.png")

	def initialize(data, name, sibling = nil)
		storage_initialize(data, nil, sibling)
		@name = name
		@size = 0
		@pvs = []
	end

	def return_space(amount, path)
		@free_node = VGFreeSpace.new(self, 0) unless @free_node and @free_node.amount > 0
		@free_node.amount += amount
	end

	def give(drive)
		return false unless drive.is_a? DriveEntity
		return false if drive.used

		@free_node = VGFreeSpace.new(self, 0) unless @free_node and @free_node.amount > 0
		@size += drive.size
		@free_node.amount += drive.size
		drive.used = self
		@pvs << drive
	end

	def cleanup
		arr = @pvs.clone
		arr.each{ |x| x.used = nil }
	end

	def free_icon
		FREE_ICON
	end

	def mkpart(size, free_row, name)
		VGPartitionEntity.new(self, "-#{name}", size, free_row)
	end

	def release(drive)
		@size -= drive.size
		@free_node.amount -= drive.size
		@pvs.delete drive
	end

	def text
		"#{@name} (#{@size}GiB)"
	end

	def opt_panel
		"vg_admin"
	end

	def prep_opt_panel(panel, builder)
		button = builder.get_object("vg_remove_pv")
		list = builder.get_object("vg_pvlist")

		if @pvs.size == 0
			button.hide
			list.hide
			return
		end

		button.show
		list.show

		prep_list(list)
	end

	def prep_list(list_widget)
		list = Gtk::ListStore.new(DriveEntity)
		list_widget.model = list

		@pvs.each do |pv|
			list.append[0] = pv
		end
	end
end

class TreeData < Gtk::TreeStore
	attr :root

	def initialize(treeview)
		super(StorageEntity)

		@treeview = treeview
		@treeview.model = self

		[["a", 320],["b",60],["c",150]].each do |letter, size|
			DriveEntity.new(self, "/dev/sd#{letter}", size)
		end

		@root = MountpointEntity.new(self, "/")
		MountpointEntity.new(self, "boot", @root.row)
		MountpointEntity.new(self, "home", @root.row)
		MountpointEntity.new(self, "var", @root.row)

		text_cell = Gtk::CellRendererText.new
		icon_cell = Gtk::CellRendererPixbuf.new
		col = Gtk::TreeViewColumn.new
		col.pack_start(icon_cell, false)
		col.pack_start(text_cell, false)
		col.set_cell_data_func(text_cell) do |tvc, cell, model, iter|
			cell.text = iter[0].text
			cell.foreground = iter[0].text_fg 
		end

		col.set_cell_data_func(icon_cell) do |tvc, cell, model, iter|
			cell.pixbuf = iter[0].icon
		end
		@treeview.append_column(col)
	end
end

class Partitioner
	def initialize(ui_file)
		@builder = Gtk::Builder.new

		@builder.add(ui_file)
		@builder.connect_signals{ |signal| method(signal) }

		@window = @builder.get_object('partitioning_window')
		@window.set_default_size(600,600)

		@trees = @builder.get_object('trees')
		@treedata = TreeData.new(@trees)

		@panels = @builder.get_object('panels')
		@cur_panel = nil

		@vg_dialog = @builder.get_object('vg_name_dialog')
		@vg_dialog_namebox = @builder.get_object('vg_name_entry')

		@trees.enable_model_drag_source(Gdk::Window::BUTTON1_MASK,
																		[['text/plain', Gtk::Drag::TARGET_SAME_WIDGET, 0]],
																		Gdk::DragContext::ACTION_DEFAULT | Gdk::DragContext::ACTION_MOVE)
		@trees.enable_model_drag_dest([['text/plain', Gtk::Drag::TARGET_SAME_WIDGET, 0]],
																	Gdk::DragContext::ACTION_DEFAULT | Gdk::DragContext::ACTION_MOVE)

		text_cell = Gtk::CellRendererText.new
		col = Gtk::TreeViewColumn.new
		col.pack_start(text_cell, false)
		col.set_cell_data_func(text_cell) do |tvc, cell, model, iter|
			cell.text = iter[0].text
		end

		@vg_list = @builder.get_object("vg_pvlist")
		@vg_list.append_column(col)
	end

	def do_drag(treeview, context, x, y, selection, info, timestamp)
		#context.drag_reply(false, timestamp) unless do_drag_maybe(treeview, x, y)
		do_drag_maybe(treeview, x, y)
		hilight
	end

	def do_drag_maybe(treeview, x, y)
		target_row = treeview.get_dest_row_at_pos(x,y)

		return false unless target_row
		return @treedata.get_iter(target_row.first)[0].give(cur_entity)
	end

	def quit
		Gtk.main_quit
	end

	def finish
		quit
	end

	def cur_entity
		arr = @treedata.get_iter(@trees.cursor[0])
		return nil unless arr
		return arr[0]
	end

	def cur_pv
		arr = @vg_list.model.get_iter(@vg_list.cursor[0])
		return nil unless arr
		return arr[0]
	end

	def act
		@panel_entity.act(@cur_panel, @builder)
	end

	def hilight
		@panels.remove(@cur_panel) if @cur_panel

		@panel_entity = cur_entity
		unless @panel_entity and @panel_entity.opt_panel
			@panel_entity = nil
			@cur_panel = nil
			return 
		end

		@cur_panel = @builder.get_object(@panel_entity.opt_panel)

		@panel_entity.prep_opt_panel(@cur_panel, @builder)

		@panels.pack_start(@cur_panel, false, false, 0)
		@panels.reorder_child(@cur_panel, 2)
	end

	def unmount
		@panel_entity.mounts.used = nil
		hilight
	end

	def create_vg
		if @vg_dialog.run == 1
			VolGroupEntity.new(@treedata, @vg_dialog_namebox.text, @treedata.root.row)
		end

		@vg_dialog.hide
	end

	def vg_remove_pv
		cur_pv.used = nil
		hilight
	end

	def vg_delete
		@panel_entity.drop
		hilight
	end

	def go_to_mountpoint
		iter = @treedata.get_iter(@panel_entity.row.path)

		until iter[0].mounts or iter.parent.nil?
			iter = iter.parent
		end

		@trees.set_cursor(iter.path, nil, false)
	end

	def delete_part
		@panel_entity.drop
		hilight
	end

	def disconnect_part
		@panel_entity.used = nil
		hilight
	end
end

Partitioner.new('ui.glade')
Gtk.main
